#ifndef PYIU_DOCSSIZEOF_H
#define PYIU_DOCSSIZEOF_H

#define PY_SSIZE_T_CLEAN
#include <Python.h>

PyDoc_STRVAR(PYIU_sizeof_doc,
    "__sizeof__(/)\n"
    "--\n\n"
    "Returns size in memory, in bytes.\n"
);

#endif
